@@ -1,8 +1,8 @@
 /* nops.c - Overlay to filter idempotent operations */
-/* $OpenLDAP: pkg/ldap/contrib/slapd-modules/nops/nops.c,v 1.1.2.5 2010/04/13 20:22:27 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>. 
  *
- * Copyright 2008-2010 The OpenLDAP Foundation.
+ * Copyright 2008-2024 The OpenLDAP Foundation.
  * Copyright 2008 Emmanuel Dreyfus.
  * All rights reserved.
  *
@@ -29,7 +29,7 @@
 
 #include "lutil.h"
 #include "slap.h"
-#include "config.h"
+#include "slap-config.h"
 
 static ConfigDriver nops_cf_gen;
 
@@ -130,8 +130,8 @@ nops_modify( Operation *op, SlapReply *rs )
 			continue;
 
 		/* This is a nop, remove it */
-		Debug(LDAP_DEBUG_TRACE, "removing nop on %s%s%s",
-			a->a_desc->ad_cname.bv_val, "", "");
+		Debug(LDAP_DEBUG_TRACE, "removing nop on %s",
+			a->a_desc->ad_cname.bv_val );
 
 		nops_rm_mod(&op->orm_modlist, mc);
 	}
@@ -163,6 +163,7 @@ static
 int
 nops_initialize( void ) {
 	nops_ovl.on_bi.bi_type = "nops";
+	nops_ovl.on_bi.bi_flags = SLAPO_BFLAG_SINGLE;
 	nops_ovl.on_bi.bi_op_modify = nops_modify;
 	return overlay_register( &nops_ovl );
 }
