@@ -2,7 +2,7 @@
 /* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2001-2018 The OpenLDAP Foundation.
+ * Copyright 2001-2024 The OpenLDAP Foundation.
  * Portions Copyright 2001-2003 Pierangelo Masarati.
  * All rights reserved.
  *
@@ -52,8 +52,7 @@ monitor_subsys_backend_init(
 			"monitor_subsys_backend_init: "
 			"unable to get "
 			"\"" SLAPD_MONITOR_DATABASE_NAME "\" "
-			"subsystem\n",
-			0, 0, 0 );
+			"subsystem\n" );
 		return -1;
 	}
 
@@ -61,7 +60,7 @@ monitor_subsys_backend_init(
 		Debug( LDAP_DEBUG_ANY,
 			"monitor_subsys_backend_init: "
 			"unable to get entry \"%s\"\n",
-			ms->mss_ndn.bv_val, 0, 0 );
+			ms->mss_ndn.bv_val );
 		return( -1 );
 	}
 
@@ -89,7 +88,7 @@ monitor_subsys_backend_init(
 			Debug( LDAP_DEBUG_ANY,
 				"monitor_subsys_backend_init: "
 				"unable to create entry \"cn=Backend %d,%s\"\n",
-				i, ms->mss_ndn.bv_val, 0 );
+				i, ms->mss_ndn.bv_val );
 			return( -1 );
 		}
 		
@@ -145,7 +144,7 @@ monitor_subsys_backend_init(
 				"monitor_subsys_backend_init: "
 				"unable to add entry \"cn=Backend %d,%s\"\n",
 				i,
-			       	ms->mss_ndn.bv_val, 0 );
+					ms->mss_ndn.bv_val );
 			return( -1 );
 		}
 
