@@ -1,8 +1,8 @@
 /* limits.c - routines to handle regex-based size and time limits */
-/* $OpenLDAP: pkg/ldap/servers/slapd/limits.c,v 1.73.2.12 2010/04/13 20:23:16 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2010 The OpenLDAP Foundation.
+ * Copyright 1998-2024 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -175,7 +175,7 @@ limits_get(
 		case SLAP_LIMITS_USERS:
 		found_nodn:
 			Debug( LDAP_DEBUG_TRACE, "<== limits_get: type=%s match=%s\n",
-				dn_source[isthis], limits2str( style ), 0 );
+				dn_source[isthis], limits2str( style ) );
 		found_any:
 			*limit = &lm[0]->lm_limits;
 			return( 0 );
@@ -328,8 +328,8 @@ limits_parse(
 	if ( argc < 3 ) {
 		Debug( LDAP_DEBUG_ANY,
 			"%s : line %d: missing arg(s) in "
-			"\"limits <pattern> <limits>\" line.\n%s",
-			fname, lineno, "" );
+			"\"limits <pattern> <limits>\" line.\n",
+			fname, lineno );
 		return( -1 );
 	}
 
@@ -413,7 +413,7 @@ limits_parse(
 					Debug( LDAP_DEBUG_ANY,
 						"%s : line %d: deprecated \"one\" style "
 						"\"limits <pattern> <limits>\" line; "
-						"use \"onelevel\" instead.\n", fname, lineno, 0 );
+						"use \"onelevel\" instead.\n", fname, lineno );
 				}
 
 			} else if ( STRSTART( pattern, "sub" ) ) {
@@ -426,7 +426,7 @@ limits_parse(
 					Debug( LDAP_DEBUG_ANY,
 						"%s : line %d: deprecated \"sub\" style "
 						"\"limits <pattern> <limits>\" line; "
-						"use \"subtree\" instead.\n", fname, lineno, 0 );
+						"use \"subtree\" instead.\n", fname, lineno );
 				}
 
 			} else if ( STRSTART( pattern, "children" ) ) {
@@ -546,7 +546,7 @@ no_ad:;
 				"\"group[/objectClass[/attributeType]]"
 				"=<pattern>\" in "
 				"\"limits <pattern> <limits>\" line.\n",
-				fname, lineno, 0 );
+				fname, lineno );
 			return( -1 );
 		}
 
@@ -619,7 +619,7 @@ no_ad:;
 		Debug( LDAP_DEBUG_ANY,
 			"%s : line %d: unable to add limit in "
 			"\"limits <pattern> <limits>\" line.\n",
-		fname, lineno, 0 );
+		fname, lineno );
 	}
 
 	return( rc );
