@@ -1,8 +1,8 @@
 /* rwm.c - rewrite/remap operations */
-/* $OpenLDAP: pkg/ldap/servers/slapd/overlays/rwm.c,v 1.70.2.36 2010/06/10 17:37:40 quanah Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2003-2010 The OpenLDAP Foundation.
+ * Copyright 2003-2024 The OpenLDAP Foundation.
  * Portions Copyright 2003 Pierangelo Masarati.
  * All rights reserved.
  *
@@ -24,7 +24,7 @@
 #include <ac/string.h>
 
 #include "slap.h"
-#include "config.h"
+#include "slap-config.h"
 #include "lutil.h"
 #include "rwm.h"
 
@@ -34,6 +34,8 @@ typedef struct rwm_op_state {
 	struct berval ro_ndn;
 	struct berval r_dn;
 	struct berval r_ndn;
+	struct berval rx_dn;
+	struct berval rx_ndn;
 	AttributeName *mapped_attrs;
 	OpRequest o_request;
 } rwm_op_state;
@@ -52,30 +54,46 @@ rwm_send_entry( Operation *op, SlapReply *rs );
 static void
 rwm_op_rollback( Operation *op, SlapReply *rs, rwm_op_state *ros )
 {
-	if ( !BER_BVISNULL( &ros->ro_dn ) ) {
-		op->o_req_dn = ros->ro_dn;
-	}
-	if ( !BER_BVISNULL( &ros->ro_ndn ) ) {
-		op->o_req_ndn = ros->ro_ndn;
-	}
+	/* in case of successful extended operation cleanup
+	 * gets called *after* (ITS#6632); this hack counts
+	 * on others to cleanup our o_req_dn/o_req_ndn,
+	 * while we cleanup theirs. */
+	if ( ros->r_tag == LDAP_REQ_EXTENDED && rs->sr_err == LDAP_SUCCESS ) {
+		if ( !BER_BVISNULL( &ros->rx_dn ) ) {
+			ch_free( ros->rx_dn.bv_val );
+		}
+		if ( !BER_BVISNULL( &ros->rx_ndn ) ) {
+			ch_free( ros->rx_ndn.bv_val );
+		}
 
-	if ( !BER_BVISNULL( &ros->r_dn )
-		&& ros->r_dn.bv_val != ros->ro_dn.bv_val )
-	{
-		assert( ros->r_dn.bv_val != ros->r_ndn.bv_val );
-		ch_free( ros->r_dn.bv_val );
-		BER_BVZERO( &ros->r_dn );
-	}
+	} else {
+		if ( !BER_BVISNULL( &ros->ro_dn ) ) {
+			op->o_req_dn = ros->ro_dn;
+		}
+		if ( !BER_BVISNULL( &ros->ro_ndn ) ) {
+			op->o_req_ndn = ros->ro_ndn;
+		}
 
-	if ( !BER_BVISNULL( &ros->r_ndn )
-		&& ros->r_ndn.bv_val != ros->ro_ndn.bv_val )
-	{
-		ch_free( ros->r_ndn.bv_val );
-		BER_BVZERO( &ros->r_ndn );
+		if ( !BER_BVISNULL( &ros->r_dn )
+			&& ros->r_dn.bv_val != ros->ro_dn.bv_val )
+		{
+			assert( ros->r_dn.bv_val != ros->r_ndn.bv_val );
+			ch_free( ros->r_dn.bv_val );
+		}
+
+		if ( !BER_BVISNULL( &ros->r_ndn )
+			&& ros->r_ndn.bv_val != ros->ro_ndn.bv_val )
+		{
+			ch_free( ros->r_ndn.bv_val );
+		}
 	}
 
+	BER_BVZERO( &ros->r_dn );
+	BER_BVZERO( &ros->r_ndn );
 	BER_BVZERO( &ros->ro_dn );
 	BER_BVZERO( &ros->ro_ndn );
+	BER_BVZERO( &ros->rx_dn );
+	BER_BVZERO( &ros->rx_ndn );
 
 	switch( ros->r_tag ) {
 	case LDAP_REQ_COMPARE:
@@ -89,10 +107,12 @@ rwm_op_rollback( Operation *op, SlapReply *rs, rwm_op_state *ros )
 		break;
 	case LDAP_REQ_MODRDN:
 		if ( op->orr_newSup != ros->orr_newSup ) {
-			ch_free( op->orr_newSup->bv_val );
-			ch_free( op->orr_nnewSup->bv_val );
-			op->o_tmpfree( op->orr_newSup, op->o_tmpmemctx );
-			op->o_tmpfree( op->orr_nnewSup, op->o_tmpmemctx );
+			if ( op->orr_newSup ) {
+				ch_free( op->orr_newSup->bv_val );
+				ch_free( op->orr_nnewSup->bv_val );
+				op->o_tmpfree( op->orr_newSup, op->o_tmpmemctx );
+				op->o_tmpfree( op->orr_nnewSup, op->o_tmpmemctx );
+			}
 			op->orr_newSup = ros->orr_newSup;
 			op->orr_nnewSup = ros->orr_nnewSup;
 		}
@@ -105,11 +125,15 @@ rwm_op_rollback( Operation *op, SlapReply *rs, rwm_op_state *ros )
 		break;
 	case LDAP_REQ_SEARCH:
 		op->o_tmpfree( ros->mapped_attrs, op->o_tmpmemctx );
-		filter_free_x( op, op->ors_filter, 1 );
-		op->o_tmpfree( op->ors_filterstr.bv_val, op->o_tmpmemctx );
 		op->ors_attrs = ros->ors_attrs;
-		op->ors_filter = ros->ors_filter;
-		op->ors_filterstr = ros->ors_filterstr;
+		if ( op->ors_filter != ros->ors_filter ) {
+			filter_free_x( op, op->ors_filter, 1 );
+			op->ors_filter = ros->ors_filter;
+		}
+		if ( op->ors_filterstr.bv_val != ros->ors_filterstr.bv_val ) {
+			op->o_tmpfree( op->ors_filterstr.bv_val, op->o_tmpmemctx );
+			op->ors_filterstr = ros->ors_filterstr;
+		}
 		break;
 	case LDAP_REQ_EXTENDED:
 		if ( op->ore_reqdata != ros->ore_reqdata ) {
@@ -154,11 +178,11 @@ rwm_op_cleanup( Operation *op, SlapReply *rs )
 }
 
 static rwm_op_cb *
-rwm_callback_get( Operation *op, SlapReply *rs )
+rwm_callback_get( Operation *op )
 {
-	rwm_op_cb	*roc = NULL;
+	rwm_op_cb	*roc;
 
-	roc = op->o_tmpalloc( sizeof( struct rwm_op_cb ), op->o_tmpmemctx );
+	roc = op->o_tmpcalloc( 1, sizeof( struct rwm_op_cb ), op->o_tmpmemctx );
 	roc->cb.sc_cleanup = rwm_op_cleanup;
 	roc->cb.sc_response = NULL;
 	roc->cb.sc_next = op->o_callback;
@@ -166,9 +190,12 @@ rwm_callback_get( Operation *op, SlapReply *rs )
 	roc->ros.r_tag = op->o_tag;
 	roc->ros.ro_dn = op->o_req_dn;
 	roc->ros.ro_ndn = op->o_req_ndn;
-	roc->ros.o_request = op->o_request;
 	BER_BVZERO( &roc->ros.r_dn );
 	BER_BVZERO( &roc->ros.r_ndn );
+	BER_BVZERO( &roc->ros.rx_dn );
+	BER_BVZERO( &roc->ros.rx_ndn );
+	roc->ros.mapped_attrs = NULL;
+	roc->ros.o_request = op->o_request;
 
 	return roc;
 }
@@ -228,6 +255,11 @@ rwm_op_dn_massage( Operation *op, SlapReply *rs, void *cookie,
 	assert( BER_BVISNULL( &ros->r_ndn ) );
 	ros->r_ndn = ndn;
 
+	if ( ros->r_tag == LDAP_REQ_EXTENDED ) {
+		ros->rx_dn = ros->r_dn;
+		ros->rx_ndn = ros->r_ndn;
+	}
+
 	return LDAP_SUCCESS;
 }
 
@@ -244,7 +276,7 @@ rwm_op_add( Operation *op, SlapReply *rs )
 	char			*olddn = op->o_req_dn.bv_val;
 	int			isupdate;
 
-	rwm_op_cb		*roc = rwm_callback_get( op, rs );
+	rwm_op_cb		*roc = rwm_callback_get( op );
 
 	rc = rwm_op_dn_massage( op, rs, "addDN", &roc->ros );
 	if ( rc != LDAP_SUCCESS ) {
@@ -387,7 +419,7 @@ rwm_op_bind( Operation *op, SlapReply *rs )
 	slap_overinst		*on = (slap_overinst *) op->o_bd->bd_info;
 	int			rc;
 
-	rwm_op_cb		*roc = rwm_callback_get( op, rs );
+	rwm_op_cb		*roc = rwm_callback_get( op );
 
 	rc = rwm_op_dn_massage( op, rs, "bindDN", &roc->ros );
 	if ( rc != LDAP_SUCCESS ) {
@@ -423,7 +455,7 @@ rwm_op_compare( Operation *op, SlapReply *rs )
 	int			rc;
 	struct berval		mapped_vals[2] = { BER_BVNULL, BER_BVNULL };
 
-	rwm_op_cb		*roc = rwm_callback_get( op, rs );
+	rwm_op_cb		*roc = rwm_callback_get( op );
 
 	rc = rwm_op_dn_massage( op, rs, "compareDN", &roc->ros );
 	if ( rc != LDAP_SUCCESS ) {
@@ -508,7 +540,7 @@ rwm_op_delete( Operation *op, SlapReply *rs )
 	slap_overinst		*on = (slap_overinst *) op->o_bd->bd_info;
 	int			rc;
 
-	rwm_op_cb		*roc = rwm_callback_get( op, rs );
+	rwm_op_cb		*roc = rwm_callback_get( op );
 
 	rc = rwm_op_dn_massage( op, rs, "deleteDN", &roc->ros );
 	if ( rc != LDAP_SUCCESS ) {
@@ -533,7 +565,7 @@ rwm_op_modify( Operation *op, SlapReply *rs )
 	Modifications		**mlp;
 	int			rc;
 
-	rwm_op_cb		*roc = rwm_callback_get( op, rs );
+	rwm_op_cb		*roc = rwm_callback_get( op );
 
 	rc = rwm_op_dn_massage( op, rs, "modifyDN", &roc->ros );
 	if ( rc != LDAP_SUCCESS ) {
@@ -575,7 +607,7 @@ rwm_op_modify( Operation *op, SlapReply *rs )
 					&mapping, RWM_MAP );
 			if ( drop_missing || ( mapping != NULL && BER_BVISNULL( &mapping->m_dst ) ) )
 			{
-				goto cleanup_mod;
+				goto skip_mod;
 			}
 		}
 
@@ -670,6 +702,10 @@ next_mod:;
 		mlp = &ml->sml_next;
 		continue;
 
+skip_mod:;
+		*mlp = (*mlp)->sml_next;
+		continue;
+
 cleanup_mod:;
 		ml = *mlp;
 		*mlp = (*mlp)->sml_next;
@@ -692,7 +728,7 @@ rwm_op_modrdn( Operation *op, SlapReply *rs )
 	int			rc;
 	dncookie		dc;
 
-	rwm_op_cb		*roc = rwm_callback_get( op, rs );
+	rwm_op_cb		*roc = rwm_callback_get( op );
 
 	if ( op->orr_newSup ) {
 		struct berval	nnewSup = BER_BVNULL;
@@ -836,12 +872,7 @@ rwm_entry_get_rw( Operation *op, struct berval *ndn,
 	ObjectClass *oc, AttributeDescription *at, int rw, Entry **ep )
 {
 	slap_overinst		*on = (slap_overinst *) op->o_bd->bd_info;
-	struct ldaprwmap	*rwmap = 
-			(struct ldaprwmap *)on->on_bi.bi_private;
-
 	int			rc;
-	dncookie		dc;
-
 	BackendDB		db;
 	Operation		op2;
 	SlapReply		rs = { REP_SEARCH };
@@ -886,9 +917,16 @@ rwm_entry_get_rw( Operation *op, struct berval *ndn,
 		/* duplicate & release */
 		op2.o_bd->bd_info = (BackendInfo *)on;
 		rc = rwm_send_entry( &op2, &rs );
+		RS_ASSERT( rs.sr_flags & REP_ENTRY_MUSTFLUSH );
 		if ( rc == SLAP_CB_CONTINUE ) {
 			*ep = rs.sr_entry;
 			rc = LDAP_SUCCESS;
+		} else {
+			assert( rc != LDAP_SUCCESS && rs.sr_entry == *ep );
+			*ep = NULL;
+			op2.o_bd->bd_info = (BackendInfo *)on->on_info;
+			be_entry_release_r( &op2, rs.sr_entry );
+			op2.o_bd->bd_info = (BackendInfo *)on;
 		}
 	}
 
@@ -916,7 +954,7 @@ rwm_op_search( Operation *op, SlapReply *rs )
 
 	char			*text = NULL;
 
-	rwm_op_cb		*roc = rwm_callback_get( op, rs );
+	rwm_op_cb		*roc = rwm_callback_get( op );
 
 	rc = rewrite_session_var_set( rwmap->rwm_rw, op->o_conn,
 		"searchFilter", op->ors_filterstr.bv_val );
@@ -1035,7 +1073,7 @@ rwm_exop_passwd( Operation *op, SlapReply *rs )
 		ber_dupbv_x( &op->o_req_ndn, &op->o_ndn, op->o_tmpmemctx );
 	}
 
-	roc = rwm_callback_get( op, rs );
+	roc = rwm_callback_get( op );
 
 	rc = rwm_op_dn_massage( op, rs, "extendedDN", &roc->ros );
 	if ( rc != LDAP_SUCCESS ) {
@@ -1107,7 +1145,7 @@ rwm_extended( Operation *op, SlapReply *rs )
 		}
 	}
 
-	roc = rwm_callback_get( op, rs );
+	roc = rwm_callback_get( op );
 
 	rc = rwm_op_dn_massage( op, rs, "extendedDN", &roc->ros );
 	if ( rc != LDAP_SUCCESS ) {
@@ -1122,7 +1160,7 @@ rwm_extended( Operation *op, SlapReply *rs )
 	return SLAP_CB_CONTINUE;
 }
 
-static int
+static void
 rwm_matched( Operation *op, SlapReply *rs )
 {
 	slap_overinst		*on = (slap_overinst *) op->o_bd->bd_info;
@@ -1134,7 +1172,7 @@ rwm_matched( Operation *op, SlapReply *rs )
 	int			rc;
 
 	if ( rs->sr_matched == NULL ) {
-		return SLAP_CB_CONTINUE;
+		return;
 	}
 
 	dc.rwmap = rwmap;
@@ -1147,10 +1185,8 @@ rwm_matched( Operation *op, SlapReply *rs )
 	if ( rc != LDAP_SUCCESS ) {
 		rs->sr_err = rc;
 		rs->sr_text = "Rewrite error";
-		return 1;
-	}
 
-	if ( mdn.bv_val != dn.bv_val ) {
+	} else if ( mdn.bv_val != dn.bv_val ) {
 		if ( rs->sr_flags & REP_MATCHED_MUSTBEFREED ) {
 			ch_free( (void *)rs->sr_matched );
 
@@ -1159,8 +1195,6 @@ rwm_matched( Operation *op, SlapReply *rs )
 		}
 		rs->sr_matched = mdn.bv_val;
 	}
-	
-	return SLAP_CB_CONTINUE;
 }
 
 static int
@@ -1249,7 +1283,13 @@ rwm_attrs( Operation *op, SlapReply *rs, Attribute** a_first, int stripEntryDN )
 								NULL );
 
 							if ( rc != LDAP_SUCCESS ) {
-								BER_BVZERO( &(*ap)->a_nvals[i] );
+								/* FIXME: this is wrong, putting a non-normalized value
+								 * into nvals. But when a proxy sends us bogus data,
+								 * we still need to give it to the client, even if it
+								 * violates the syntax. I.e., we don't want to silently
+								 * drop things and trigger an apparent data loss.
+								 */
+								ber_dupbv( &(*ap)->a_nvals[i], &(*ap)->a_vals[i] );
 							}
 						}
 						BER_BVZERO( &(*ap)->a_nvals[i] );
@@ -1431,6 +1471,7 @@ cleanup_attr:;
 	return 0;
 }
 
+/* Should return SLAP_CB_CONTINUE or failure, never LDAP_SUCCESS. */
 static int
 rwm_send_entry( Operation *op, SlapReply *rs )
 {
@@ -1439,7 +1480,6 @@ rwm_send_entry( Operation *op, SlapReply *rs )
 			(struct ldaprwmap *)on->on_bi.bi_private;
 
 	Entry			*e = NULL;
-	slap_mask_t		flags;
 	struct berval		dn = BER_BVNULL,
 				ndn = BER_BVNULL;
 	dncookie		dc;
@@ -1456,7 +1496,6 @@ rwm_send_entry( Operation *op, SlapReply *rs )
 	dc.ctx = "searchEntryDN";
 
 	e = rs->sr_entry;
-	flags = rs->sr_flags;
 	if ( !( rs->sr_flags & REP_ENTRY_MODIFIABLE ) ) {
 		/* FIXME: all we need to duplicate are:
 		 * - dn
@@ -1464,15 +1503,17 @@ rwm_send_entry( Operation *op, SlapReply *rs )
 		 * - attributes that are requested
 		 * - no values if attrsonly is set
 		 */
-
 		e = entry_dup( e );
 		if ( e == NULL ) {
 			rc = LDAP_NO_MEMORY;
 			goto fail;
 		}
-
-		flags &= ~REP_ENTRY_MUSTRELEASE;
-		flags |= ( REP_ENTRY_MODIFIABLE | REP_ENTRY_MUSTBEFREED );
+	} else if ( rs->sr_flags & REP_ENTRY_MUSTRELEASE ) {
+		/* ITS#6423: REP_ENTRY_MUSTRELEASE incompatible
+		 * with REP_ENTRY_MODIFIABLE */
+		RS_ASSERT( 0 );
+		rc = 1;
+		goto fail;
 	}
 
 	/*
@@ -1504,22 +1545,21 @@ rwm_send_entry( Operation *op, SlapReply *rs )
 	 * to return, and remap them accordingly */
 	(void)rwm_attrs( op, rs, &e->e_attrs, 1 );
 
-	if ( rs->sr_flags & REP_ENTRY_MUSTRELEASE ) {
-		/* ITS#6423: REP_ENTRY_MUSTRELEASE incompatible
-		 * with REP_ENTRY_MODIFIABLE */
-		if ( rs->sr_entry == e ) {
-			rc = 1;
-			goto fail;
+	if ( e != rs->sr_entry ) {
+		/* Reimplementing rs_replace_entry(), I suppose to
+		 * bypass our own dubious rwm_entry_release_rw() */
+		if ( rs->sr_flags & REP_ENTRY_MUSTRELEASE ) {
+			rs->sr_flags ^= REP_ENTRY_MUSTRELEASE;
+			op->o_bd->bd_info = (BackendInfo *)on->on_info;
+			be_entry_release_r( op, rs->sr_entry );
+			op->o_bd->bd_info = (BackendInfo *)on;
+		} else if ( rs->sr_flags & REP_ENTRY_MUSTBEFREED ) {
+			entry_free( rs->sr_entry );
 		}
-
-		op->o_bd->bd_info = (BackendInfo *)on->on_info;
-		be_entry_release_r( op, rs->sr_entry );
-		op->o_bd->bd_info = (BackendInfo *)on;
+		rs->sr_entry = e;
+		rs->sr_flags |= REP_ENTRY_MODIFIABLE | REP_ENTRY_MUSTBEFREED;
 	}
 
-	rs->sr_entry = e;
-	rs->sr_flags = flags;
-
 	return SLAP_CB_CONTINUE;
 
 fail:;
@@ -1631,7 +1671,7 @@ rwm_suffixmassage_config(
 				       " <massaged suffix>\" without "
 				       "<suffix> part requires database "
 				       "suffix be defined first.\n",
-				fname, lineno, 0 );
+				fname, lineno );
 			return 1;
 		}
 		bvnc = be->be_suffix[ 0 ];
@@ -1645,7 +1685,7 @@ rwm_suffixmassage_config(
  		Debug( LDAP_DEBUG_ANY, "%s: line %d: syntax is"
 			       " \"suffixMassage [<suffix>]"
 			       " <massaged suffix>\"\n",
-			fname, lineno, 0 );
+			fname, lineno );
 		return 1;
 	}
 
@@ -1729,20 +1769,18 @@ rwm_response( Operation *op, SlapReply *rs )
 			dc.rs = NULL; 
 			dc.ctx = "referralDN";
 			rc = rwm_referral_result_rewrite( &dc, rs->sr_ref );
+			/* FIXME: impossible, so far */
 			if ( rc != LDAP_SUCCESS ) {
-				rc = 1;
+				rs->sr_err = rc;
 				break;
 			}
 		}
-		rc = rwm_matched( op, rs );
-		break;
 
-	default:
-		rc = SLAP_CB_CONTINUE;
+		rwm_matched( op, rs );
 		break;
 	}
 
-	return rc;
+	return SLAP_CB_CONTINUE;
 }
 
 static int
@@ -1778,7 +1816,7 @@ rwm_db_config(
 		if ( argc != 2 ) {
 			Debug( LDAP_DEBUG_ANY,
 		"%s: line %d: \"t-f-support {no|yes|discover}\" needs 1 argument.\n",
-					fname, lineno, 0 );
+					fname, lineno );
 			return( 1 );
 		}
 
@@ -1793,7 +1831,7 @@ rwm_db_config(
 			Debug( LDAP_DEBUG_ANY,
 		"%s: line %d: \"discover\" not supported yet "
 		"in \"t-f-support {no|yes|discover}\".\n",
-					fname, lineno, 0 );
+					fname, lineno );
 			return( 1 );
 #if 0
 			rwmap->rwm_flags |= RWM_F_SUPPORT_T_F_DISCOVER;
@@ -1810,7 +1848,7 @@ rwm_db_config(
 		if ( argc !=2 ) { 
 			Debug( LDAP_DEBUG_ANY,
 		"%s: line %d: \"normalize-mapped-attrs {no|yes}\" needs 1 argument.\n",
-					fname, lineno, 0 );
+					fname, lineno );
 			return( 1 );
 		}
 
@@ -1879,6 +1917,7 @@ static ConfigTable rwmcfg[] = {
 		2, 2, 0, ARG_MAGIC|RWM_CF_T_F_SUPPORT, rwm_cf_gen,
 		"( OLcfgOvAt:16.2 NAME 'olcRwmTFSupport' "
 			"DESC 'Absolute filters support' "
+			"EQUALITY caseIgnoreMatch "
 			"SYNTAX OMsDirectoryString "
 			"SINGLE-VALUE )",
 		NULL, NULL },
@@ -1896,6 +1935,7 @@ static ConfigTable rwmcfg[] = {
 		2, 2, 0, ARG_MAGIC|ARG_ON_OFF|RWM_CF_NORMALIZE_MAPPED, rwm_cf_gen,
 		"( OLcfgOvAt:16.4 NAME 'olcRwmNormalizeMapped' "
 			"DESC 'Normalize mapped attributes/objectClasses' "
+			"EQUALITY booleanMatch "
 			"SYNTAX OMsBoolean "
 			"SINGLE-VALUE )",
 		NULL, NULL },
@@ -1904,6 +1944,7 @@ static ConfigTable rwmcfg[] = {
 		2, 2, 0, ARG_MAGIC|ARG_ON_OFF|RWM_CF_DROP_UNREQUESTED, rwm_cf_gen,
 		"( OLcfgOvAt:16.5 NAME 'olcRwmDropUnrequested' "
 			"DESC 'Drop unrequested attributes' "
+			"EQUALITY booleanMatch "
 			"SYNTAX OMsBoolean "
 			"SINGLE-VALUE )",
 		NULL, NULL },
@@ -1920,52 +1961,13 @@ static ConfigOCs rwmocs[] = {
 			"olcRwmRewrite $ "
 			"olcRwmTFSupport $ "
 			"olcRwmMap $ "
-			"olcRwmNormalizeMapped "
+			"olcRwmNormalizeMapped $ "
+			"olcRwmDropUnrequested"
 			") )",
 		Cft_Overlay, rwmcfg, NULL, NULL },
 	{ NULL, 0, NULL }
 };
 
-static void
-slap_bv_x_ordered_unparse( BerVarray in, BerVarray *out )
-{
-	int		i;
-	BerVarray	bva = NULL;
-	char		ibuf[32], *ptr;
-	struct berval	idx;
-
-	assert( in != NULL );
-
-	for ( i = 0; !BER_BVISNULL( &in[i] ); i++ )
-		/* count'em */ ;
-
-	if ( i == 0 ) {
-		return;
-	}
-
-	idx.bv_val = ibuf;
-
-	bva = ch_malloc( ( i + 1 ) * sizeof(struct berval) );
-	BER_BVZERO( &bva[ 0 ] );
-
-	for ( i = 0; !BER_BVISNULL( &in[i] ); i++ ) {
-		idx.bv_len = snprintf( idx.bv_val, sizeof( ibuf ), "{%d}", i );
-		if ( idx.bv_len >= sizeof( ibuf ) ) {
-			ber_bvarray_free( bva );
-			return;
-		}
-
-		bva[i].bv_len = idx.bv_len + in[i].bv_len;
-		bva[i].bv_val = ch_malloc( bva[i].bv_len + 1 );
-		ptr = lutil_strcopy( bva[i].bv_val, ibuf );
-		ptr = lutil_strcopy( ptr, in[i].bv_val );
-		*ptr = '\0';
-		BER_BVZERO( &bva[ i + 1 ] );
-	}
-
-	*out = bva;
-}
-
 static int
 rwm_bva_add(
 	BerVarray		*bva,
@@ -2006,6 +2008,7 @@ rwm_bva_rewrite_add(
 	return rwm_bva_add( &rwmap->rwm_bva_rewrite, idx, argv );
 }
 
+#ifdef unused
 static int
 rwm_bva_map_add(
 	struct ldaprwmap	*rwmap,
@@ -2014,6 +2017,7 @@ rwm_bva_map_add(
 {
 	return rwm_bva_add( &rwmap->rwm_bva_map, idx, argv );
 }
+#endif /* unused */
 
 static int
 rwm_info_init( struct rewrite_info ** rwm_rw )
@@ -2064,10 +2068,7 @@ rwm_cf_gen( ConfigArgs *c )
 				rc = 1;
 
 			} else {
-				slap_bv_x_ordered_unparse( rwmap->rwm_bva_rewrite, &c->rvalue_vals );
-				if ( !c->rvalue_vals ) {
-					rc = 1;
-				}
+				rc = slap_bv_x_ordered_unparse( rwmap->rwm_bva_rewrite, &c->rvalue_vals );
 			}
 			break;
 
@@ -2142,8 +2143,12 @@ rwm_cf_gen( ConfigArgs *c )
 
 					ca.line = rwmap->rwm_bva_rewrite[ i ].bv_val;
 					ca.argc = 0;
-					config_fp_parse_line( &ca );
-					
+					init_config_argv( &ca );
+					config_parse_ldif( &ca );
+
+					argv0 = ca.argv[ 0 ];
+					ca.argv[ 0 ] += STRLENOF( "rwm-" );
+
 					if ( strcasecmp( ca.argv[ 0 ], "suffixmassage" ) == 0 ) {
 						rc = rwm_suffixmassage_config( &db, c->fname, c->lineno,
 							ca.argc, ca.argv );
@@ -2153,6 +2158,8 @@ rwm_cf_gen( ConfigArgs *c )
 							ca.argc, ca.argv );
 					}
 
+					ca.argv[ 0 ] = argv0;
+
 					ch_free( ca.tline );
 					ch_free( ca.argv );
 
@@ -2206,7 +2213,8 @@ rwm_cf_gen( ConfigArgs *c )
 
 					ca.line = rwmap->rwm_bva_map[ cnt ].bv_val;
 					ca.argc = 0;
-					config_fp_parse_line( &ca );
+					init_config_argv( &ca );
+					config_parse_ldif( &ca );
 					
 					argv[1] = ca.argv[0];
 					argv[2] = ca.argv[1];
@@ -2221,10 +2229,10 @@ rwm_cf_gen( ConfigArgs *c )
 					/* in case of failure, restore
 					 * the existing mapping */
 					if ( rc ) {
-						avl_free( rwmap->rwm_oc.remap, rwm_mapping_dst_free );
-						avl_free( rwmap->rwm_oc.map, rwm_mapping_free );
-						avl_free( rwmap->rwm_at.remap, rwm_mapping_dst_free );
-						avl_free( rwmap->rwm_at.map, rwm_mapping_free );
+						ldap_avl_free( rwmap->rwm_oc.remap, rwm_mapping_dst_free );
+						ldap_avl_free( rwmap->rwm_oc.map, rwm_mapping_free );
+						ldap_avl_free( rwmap->rwm_at.remap, rwm_mapping_dst_free );
+						ldap_avl_free( rwmap->rwm_at.map, rwm_mapping_free );
 						rwmap->rwm_oc = rwm_oc;
 						rwmap->rwm_at = rwm_at;
 						break;
@@ -2234,10 +2242,10 @@ rwm_cf_gen( ConfigArgs *c )
 				/* in case of success, destroy the old mapping
 				 * and eliminate the deleted one */
 				if ( rc == 0 ) {
-					avl_free( rwm_oc.remap, rwm_mapping_dst_free );
-					avl_free( rwm_oc.map, rwm_mapping_free );
-					avl_free( rwm_at.remap, rwm_mapping_dst_free );
-					avl_free( rwm_at.map, rwm_mapping_free );
+					ldap_avl_free( rwm_oc.remap, rwm_mapping_dst_free );
+					ldap_avl_free( rwm_oc.map, rwm_mapping_free );
+					ldap_avl_free( rwm_at.remap, rwm_mapping_dst_free );
+					ldap_avl_free( rwm_at.map, rwm_mapping_free );
 
 					ber_memfree( rwmap->rwm_bva_map[ c->valx ].bv_val );
 					for ( cnt = c->valx; !BER_BVISNULL( &rwmap->rwm_bva_map[ cnt ] ); cnt++ ) {
@@ -2246,10 +2254,10 @@ rwm_cf_gen( ConfigArgs *c )
 				}
 
 			} else {
-				avl_free( rwmap->rwm_oc.remap, rwm_mapping_dst_free );
-				avl_free( rwmap->rwm_oc.map, rwm_mapping_free );
-				avl_free( rwmap->rwm_at.remap, rwm_mapping_dst_free );
-				avl_free( rwmap->rwm_at.map, rwm_mapping_free );
+				ldap_avl_free( rwmap->rwm_oc.remap, rwm_mapping_dst_free );
+				ldap_avl_free( rwmap->rwm_oc.map, rwm_mapping_free );
+				ldap_avl_free( rwmap->rwm_at.remap, rwm_mapping_dst_free );
+				ldap_avl_free( rwmap->rwm_at.map, rwm_mapping_free );
 
 				rwmap->rwm_oc.remap = NULL;
 				rwmap->rwm_oc.map = NULL;
@@ -2300,7 +2308,8 @@ rwm_cf_gen( ConfigArgs *c )
 
 				ca.line = rwmap->rwm_bva_rewrite[ i ].bv_val;
 				ca.argc = 0;
-				config_fp_parse_line( &ca );
+				init_config_argv( &ca );
+				config_parse_ldif( &ca );
 
 				argv0 = ca.argv[ 0 ];
 				ca.argv[ 0 ] += STRLENOF( "rwm-" );
@@ -2350,7 +2359,8 @@ rwm_cf_gen( ConfigArgs *c )
 
 				ca.line = rwmap->rwm_bva_rewrite[ i ].bv_val;
 				ca.argc = 0;
-				config_fp_parse_line( &ca );
+				init_config_argv( &ca );
+				config_parse_ldif( &ca );
 				
 				argv0 = ca.argv[ 0 ];
 				ca.argv[ 0 ] += STRLENOF( "rwm-" );
@@ -2449,7 +2459,8 @@ rwm_cf_gen( ConfigArgs *c )
 
 				ca.line = rwmap->rwm_bva_map[ cnt ].bv_val;
 				ca.argc = 0;
-				config_fp_parse_line( &ca );
+				init_config_argv( &ca );
+				config_parse_ldif( &ca );
 
 				argv[1] = ca.argv[0];
 				argv[2] = ca.argv[1];
@@ -2482,7 +2493,8 @@ rwm_cf_gen( ConfigArgs *c )
 
 					ca.line = rwmap->rwm_bva_map[ cnt ].bv_val;
 					ca.argc = 0;
-					config_fp_parse_line( &ca );
+					init_config_argv( &ca );
+					config_parse_ldif( &ca );
 			
 					argv[1] = ca.argv[0];
 					argv[2] = ca.argv[1];
@@ -2523,10 +2535,10 @@ rwm_cf_gen( ConfigArgs *c )
 				rwmap->rwm_bva_map = tmp;
 				BER_BVZERO( &rwmap->rwm_bva_map[ cnt + 1 ] );
 
-				avl_free( rwm_oc.remap, rwm_mapping_dst_free );
-				avl_free( rwm_oc.map, rwm_mapping_free );
-				avl_free( rwm_at.remap, rwm_mapping_dst_free );
-				avl_free( rwm_at.map, rwm_mapping_free );
+				ldap_avl_free( rwm_oc.remap, rwm_mapping_dst_free );
+				ldap_avl_free( rwm_oc.map, rwm_mapping_free );
+				ldap_avl_free( rwm_at.remap, rwm_mapping_dst_free );
+				ldap_avl_free( rwm_at.map, rwm_mapping_free );
 
 				for ( ; cnt-- > c->valx; ) {
 					rwmap->rwm_bva_map[ cnt + 1 ] = rwmap->rwm_bva_map[ cnt ];
@@ -2535,10 +2547,10 @@ rwm_cf_gen( ConfigArgs *c )
 
 			} else {
 rwmmap_fail:;
-				avl_free( rwmap->rwm_oc.remap, rwm_mapping_dst_free );
-				avl_free( rwmap->rwm_oc.map, rwm_mapping_free );
-				avl_free( rwmap->rwm_at.remap, rwm_mapping_dst_free );
-				avl_free( rwmap->rwm_at.map, rwm_mapping_free );
+				ldap_avl_free( rwmap->rwm_oc.remap, rwm_mapping_dst_free );
+				ldap_avl_free( rwmap->rwm_oc.map, rwm_mapping_free );
+				ldap_avl_free( rwmap->rwm_at.remap, rwm_mapping_dst_free );
+				ldap_avl_free( rwmap->rwm_at.map, rwm_mapping_free );
 				rwmap->rwm_oc = rwm_oc;
 				rwmap->rwm_at = rwm_at;
 			}
@@ -2632,10 +2644,10 @@ rwm_db_destroy(
 				ber_bvarray_free( rwmap->rwm_bva_rewrite );
 		}
 
-		avl_free( rwmap->rwm_oc.remap, rwm_mapping_dst_free );
-		avl_free( rwmap->rwm_oc.map, rwm_mapping_free );
-		avl_free( rwmap->rwm_at.remap, rwm_mapping_dst_free );
-		avl_free( rwmap->rwm_at.map, rwm_mapping_free );
+		ldap_avl_free( rwmap->rwm_oc.remap, rwm_mapping_dst_free );
+		ldap_avl_free( rwmap->rwm_oc.map, rwm_mapping_free );
+		ldap_avl_free( rwmap->rwm_at.remap, rwm_mapping_dst_free );
+		ldap_avl_free( rwmap->rwm_at.map, rwm_mapping_free );
 		ber_bvarray_free( rwmap->rwm_bva_map );
 
 		ch_free( rwmap );
