@@ -1,8 +1,8 @@
 /* result.c - routines to send ldap results, errors, and referrals */
-/* $OpenLDAP: pkg/ldap/servers/slapd/result.c,v 1.289.2.33 2010/04/14 17:28:47 quanah Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2010 The OpenLDAP Foundation.
+ * Copyright 1998-2024 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -37,6 +37,28 @@
 
 #include "slap.h"
 
+#if SLAP_STATS_ETIME
+#define ETIME_SETUP \
+	struct timeval now; \
+	char timestr[64]; \
+	(void) gettimeofday( &now, NULL ); \
+	now.tv_sec -= op->o_time; \
+	now.tv_usec -= op->o_tusec; \
+	if ( now.tv_usec < 0 ) { \
+		--now.tv_sec; now.tv_usec += 1000000; \
+	} \
+	sprintf(timestr, "qtime=%d.%06d etime=%d.%06d", \
+		(int)op->o_qtime.tv_sec, (int)op->o_qtime.tv_usec, \
+		(int)now.tv_sec, (int)now.tv_usec);
+#define ETIME_LOGFMT	"%s "
+#define StatslogEtime(lvl,fmt,pfx,tag,err,...) \
+	Debug(lvl,fmt,pfx,tag,err,timestr,__VA_ARGS__)
+#else
+#define ETIME_SETUP
+#define ETIME_LOGFMT	""
+#define StatslogEtime	Debug
+#endif	/* SLAP_STATS_ETIME */
+
 const struct berval slap_dummy_bv = BER_BVNULL;
 
 int slap_null_cb( Operation *op, SlapReply *rs )
@@ -132,30 +154,147 @@ slap_req2res( ber_tag_t tag )
 	return tag;
 }
 
-#ifdef RS_ASSERT
-#elif 0 && defined LDAP_DEVEL /* FIXME: this should not crash. ITS#5340. */
-#define RS_ASSERT assert
-#else
-#define RS_ASSERT(cond) ((void) 0)
-#endif
+/*
+ * SlapReply debugging enabled by USE_RS_ASSERT.
+ *
+ * Disabled by default, but compiled in (but still unused) when
+ * LDAP_TEST.  #define USE_RS_ASSERT as nonzero to enable some
+ * assertions which check the SlapReply.  USE_RS_ASSERT = 2 or higher
+ * check aggressively, currently some code fail these tests.
+ *
+ * Environment variable $NO_RS_ASSERT controls how USE_RS_ASSERT handles
+ * errors.  > 0: ignore errors, 0: abort (the default), < 0: just warn.
+ *
+ * Wrap LDAP operation calls in macros SLAP_OP() & co from proto-slap.h
+ * to check the SlapReply.  contrib/slapd-tools/wrap_slap_ops converts
+ * source code to use the macros.
+ */
+#if defined(LDAP_TEST) || (defined(USE_RS_ASSERT) && (USE_RS_ASSERT))
 
-/* Set rs->sr_entry after obyeing and clearing sr_flags & REP_ENTRY_MASK. */
-void
-rs_replace_entry( Operation *op, SlapReply *rs, slap_overinst *on, Entry *e )
+int rs_suppress_assert = 0;
+
+/* RS_ASSERT() helper function */
+void rs_assert_(const char*file, unsigned line, const char*fn, const char*cond)
 {
-	slap_mask_t e_flags = rs->sr_flags & REP_ENTRY_MUSTFLUSH;
+	int no_assert = rs_suppress_assert, save_errno = errno;
+	const char *s;
 
-	if ( e_flags && rs->sr_entry != NULL ) {
-		RS_ASSERT( e_flags != REP_ENTRY_MUSTFLUSH );
-		if ( !(e_flags & REP_ENTRY_MUSTRELEASE) ) {
+	if ( no_assert >= 0 ) {
+		if ( no_assert == 0 && (s = getenv( "NO_RS_ASSERT" )) && *s ) {
+			no_assert = rs_suppress_assert = atoi( s );
+		}
+		if ( no_assert > 0 ) {
+			errno = save_errno;
+			return;
+		}
+	}
+
+#ifdef rs_assert_	/* proto-slap.h #defined away the fn parameter */
+	fprintf( stderr,"%s:%u: "  "RS_ASSERT(%s) failed.\n", file,line,cond );
+#else
+	fprintf( stderr,"%s:%u: %s: RS_ASSERT(%s) failed.\n", file,line,fn,cond );
+#endif
+	fflush( stderr );
+
+	errno = save_errno;
+	/* $NO_RS_ASSERT > 0: ignore rs_asserts, 0: abort, < 0: just warn */
+	if ( !no_assert /* from $NO_RS_ASSERT */ ) abort();
+}
+
+/* SlapReply is consistent */
+void
+(rs_assert_ok)( const SlapReply *rs )
+{
+	const slap_mask_t flags = rs->sr_flags;
+
+	if ( flags & REP_ENTRY_MASK ) {
+		RS_ASSERT( !(flags & REP_ENTRY_MUSTRELEASE)
+			|| !(flags & (REP_ENTRY_MASK ^ REP_ENTRY_MUSTRELEASE)) );
+		RS_ASSERT( rs->sr_entry != NULL );
+		RS_ASSERT( (1 << rs->sr_type) &
+			((1 << REP_SEARCH) | (1 << REP_SEARCHREF) |
+			 (1 << REP_RESULT) | (1 << REP_GLUE_RESULT)) );
+	}
+#if defined(USE_RS_ASSERT) && (USE_RS_ASSERT) > 1 /* TODO: Enable when safe */
+	if ( (flags & (REP_MATCHED_MASK | REP_REF_MASK | REP_CTRLS_MASK)) ) {
+		RS_ASSERT( !(flags & REP_MATCHED_MASK) || rs->sr_matched );
+		RS_ASSERT( !(flags & REP_CTRLS_MASK  ) || rs->sr_ctrls   );
+		/* Note: LDAP_REFERRAL + !sr_ref is OK, becomes LDAP_NO_SUCH_OBJECT */
+	}
+#if (USE_RS_ASSERT) > 2
+	if ( rs->sr_err == LDAP_SUCCESS ) {
+		RS_ASSERT( rs->sr_text == NULL );
+		RS_ASSERT( rs->sr_matched == NULL );
+	}
+#endif
+#endif
+}
+
+/* Ready for calling a new backend operation */
+void
+(rs_assert_ready)( const SlapReply *rs )
+{
+	RS_ASSERT( !rs->sr_entry   );
+#if defined(USE_RS_ASSERT) && (USE_RS_ASSERT) > 1 /* TODO: Enable when safe */
+	RS_ASSERT( !rs->sr_text    );
+	RS_ASSERT( !rs->sr_ref     );
+	RS_ASSERT( !rs->sr_matched );
+	RS_ASSERT( !rs->sr_ctrls   );
+	RS_ASSERT( !rs->sr_flags   );
+#if (USE_RS_ASSERT) > 2
+	RS_ASSERT( rs->sr_err == LDAP_SUCCESS );
+#endif
+#else
+	RS_ASSERT( !(rs->sr_flags & REP_ENTRY_MASK) );
+#endif
+}
+
+/* Backend operation done */
+void
+(rs_assert_done)( const SlapReply *rs )
+{
+#if defined(USE_RS_ASSERT) && (USE_RS_ASSERT) > 1 /* TODO: Enable when safe */
+	RS_ASSERT( !(rs->sr_flags & ~(REP_ENTRY_MODIFIABLE|REP_NO_OPERATIONALS)) );
+	rs_assert_ok( rs );
+#else
+	RS_ASSERT( !(rs->sr_flags & REP_ENTRY_MUSTFLUSH) );
+#endif
+}
+
+#endif /* LDAP_TEST || USE_RS_ASSERT */
+
+/* Reset a used SlapReply whose contents has been flushed (freed/released) */
+void
+(rs_reinit)( SlapReply *rs, slap_reply_t type )
+{
+	rs_reinit( rs, type );		/* proto-slap.h macro */
+}
+
+/* Obey and clear rs->sr_flags & REP_ENTRY_MASK.  Clear sr_entry if freed. */
+void
+rs_flush_entry( Operation *op, SlapReply *rs, slap_overinst *on )
+{
+	rs_assert_ok( rs );
+
+	if ( (rs->sr_flags & REP_ENTRY_MUSTFLUSH) && rs->sr_entry != NULL ) {
+		if ( !(rs->sr_flags & REP_ENTRY_MUSTRELEASE) ) {
 			entry_free( rs->sr_entry );
 		} else if ( on != NULL ) {
 			overlay_entry_release_ov( op, rs->sr_entry, 0, on );
 		} else {
 			be_entry_release_rw( op, rs->sr_entry, 0 );
 		}
+		rs->sr_entry = NULL;
 	}
+
 	rs->sr_flags &= ~REP_ENTRY_MASK;
+}
+
+/* Set rs->sr_entry after obeying and clearing sr_flags & REP_ENTRY_MASK. */
+void
+rs_replace_entry( Operation *op, SlapReply *rs, slap_overinst *on, Entry *e )
+{
+	rs_flush_entry( op, rs, on );
 	rs->sr_entry = e;
 }
 
@@ -165,10 +304,10 @@ rs_replace_entry( Operation *op, SlapReply *rs, slap_overinst *on, Entry *e )
  * Return nonzero if rs->sr_entry was replaced.
  */
 int
-rs_ensure_entry_modifiable( Operation *op, SlapReply *rs, slap_overinst *on )
+rs_entry2modifiable( Operation *op, SlapReply *rs, slap_overinst *on )
 {
 	if ( rs->sr_flags & REP_ENTRY_MODIFIABLE ) {
-		RS_ASSERT((rs->sr_flags & REP_ENTRY_MUSTFLUSH)==REP_ENTRY_MUSTBEFREED);
+		rs_assert_ok( rs );
 		return 0;
 	}
 	rs_replace_entry( op, rs, on, entry_dup( rs->sr_entry ));
@@ -176,6 +315,22 @@ rs_ensure_entry_modifiable( Operation *op, SlapReply *rs, slap_overinst *on )
 	return 1;
 }
 
+/* Check for any callbacks that want to be informed about being blocked
+ * on output. These callbacks are expected to leave the callback list
+ * unmodified. Their result is ignored.
+ */
+static void
+slap_writewait_play(
+	Operation *op )
+{
+	slap_callback	*sc = op->o_callback;
+
+	for ( ; sc; sc = sc->sc_next ) {
+		if ( sc->sc_writewait )
+			sc->sc_writewait( op, sc );
+	}
+}
+
 static long send_ldap_ber(
 	Operation *op,
 	BerElement *ber )
@@ -183,6 +338,8 @@ static long send_ldap_ber(
 	Connection *conn = op->o_conn;
 	ber_len_t bytes;
 	long ret = 0;
+	char *close_reason;
+	int do_resume = 0;
 
 	ber_get_option( ber, LBER_OPT_BER_BYTES_TO_WRITE, &bytes );
 
@@ -216,24 +373,9 @@ static long send_ldap_ber(
 	/* write the pdu */
 	while( 1 ) {
 		int err;
-
-		/* lock the connection */ 
-		if ( ldap_pvt_thread_mutex_trylock( &conn->c_mutex )) {
-			if ( !connection_valid(conn)) {
-				ret = 0;
-				break;
-			}
-			ldap_pvt_thread_mutex_unlock( &conn->c_write1_mutex );
-			ldap_pvt_thread_mutex_lock( &conn->c_write1_mutex );
-			if ( conn->c_writers < 0 ) {
-				ret = 0;
-				break;
-			}
-			continue;
-		}
+		char ebuf[128];
 
 		if ( ber_flush2( conn->c_sb, ber, LBER_FLUSH_FREE_NEVER ) == 0 ) {
-			ldap_pvt_thread_mutex_unlock( &conn->c_mutex );
 			ret = bytes;
 			break;
 		}
@@ -247,29 +389,49 @@ static long send_ldap_ber(
 		 */
 
 		Debug( LDAP_DEBUG_CONNS, "ber_flush2 failed errno=%d reason=\"%s\"\n",
-		    err, sock_errstr(err), 0 );
+		    err, sock_errstr(err, ebuf, sizeof(ebuf)) );
 
 		if ( err != EWOULDBLOCK && err != EAGAIN ) {
+			close_reason = "connection lost on write";
+fail:
 			conn->c_writers--;
 			conn->c_writing = 0;
 			ldap_pvt_thread_mutex_unlock( &conn->c_write1_mutex );
-			connection_closing( conn, "connection lost on write" );
-
+			ldap_pvt_thread_mutex_lock( &conn->c_mutex );
+			/* conn may have been reused by the time we get the mutex */
+			if ( op->o_connid == conn->c_connid )
+				connection_closing( conn, close_reason );
 			ldap_pvt_thread_mutex_unlock( &conn->c_mutex );
 			return -1;
 		}
 
-		/* wait for socket to be write-ready */
-		ldap_pvt_thread_mutex_lock( &conn->c_write2_mutex );
-		conn->c_writewaiter = 1;
-		slapd_set_write( conn->c_sd, 2 );
+		/* if writer is blocked and we're waiting for a pool pause,
+		 * just drop this connection.
+		 */
+		if ( ldap_pvt_thread_pool_pausing( &connection_pool ) > 0 ) {
+			close_reason = "writer blocked and pool pause pending";
+			goto fail;
+		}
 
+		/* wait for socket to be write-ready */
+		do_resume = 1;
+		conn->c_writewaiter = 1;
 		ldap_pvt_thread_mutex_unlock( &conn->c_write1_mutex );
-		ldap_pvt_thread_mutex_unlock( &conn->c_mutex );
-		ldap_pvt_thread_cond_wait( &conn->c_write2_cv, &conn->c_write2_mutex );
+		slap_writewait_play( op );
+		err = slapd_wait_writer( conn->c_sd );
 		conn->c_writewaiter = 0;
-		ldap_pvt_thread_mutex_unlock( &conn->c_write2_mutex );
 		ldap_pvt_thread_mutex_lock( &conn->c_write1_mutex );
+		/* 0 is timeout, so we close it.
+		 * -1 is an error, close it.
+		 */
+		if ( err <= 0 ) {
+			if ( err == 0 )
+				close_reason = "writetimeout";
+			else
+				close_reason = "connection lost on writewait";
+			goto fail;
+		}
+
 		if ( conn->c_writers < 0 ) {
 			ret = 0;
 			break;
@@ -278,15 +440,24 @@ static long send_ldap_ber(
 
 	conn->c_writing = 0;
 	if ( conn->c_writers < 0 ) {
+		/* shutting down, don't resume any ops */
+		do_resume = 0;
 		conn->c_writers++;
 		if ( !conn->c_writers )
 			ldap_pvt_thread_cond_signal( &conn->c_write1_cv );
 	} else {
 		conn->c_writers--;
+		/* other writers are waiting, don't resume any ops */
+		if ( conn->c_writers )
+			do_resume = 0;
 		ldap_pvt_thread_cond_signal( &conn->c_write1_cv );
 	}
 	ldap_pvt_thread_mutex_unlock( &conn->c_write1_mutex );
 
+	/* If there are no more writers, release a pending op */
+	if ( do_resume )
+		connection_write_resume( conn );
+
 	return ret;
 }
 
@@ -467,7 +638,8 @@ send_ldap_response(
 	int		rc = LDAP_SUCCESS;
 	long	bytes;
 
-	if (( rs->sr_err == SLAPD_ABANDON || op->o_abandon ) && !op->o_cancel ) {
+	/* op was actually aborted, bypass everything if client didn't Cancel */
+	if (( rs->sr_err == SLAPD_ABANDON ) && !op->o_cancel ) {
 		rc = SLAPD_ABANDON;
 		goto clean2;
 	}
@@ -479,6 +651,12 @@ send_ldap_response(
 		}
 	}
 
+	/* op completed, connection aborted, bypass sending response */
+	if ( op->o_abandon && !op->o_cancel ) {
+		rc = SLAPD_ABANDON;
+		goto clean2;
+	}
+
 #ifdef LDAP_CONNECTIONLESS
 	if (op->o_conn && op->o_conn->c_is_udp)
 		ber = op->o_res_ber;
@@ -499,8 +677,7 @@ send_ldap_response(
 
 	if( rs->sr_ref ) {
 		Debug( LDAP_DEBUG_ARGS, "send_ldap_response: ref=\"%s\"\n",
-			rs->sr_ref[0].bv_val ? rs->sr_ref[0].bv_val : "NULL",
-			NULL, NULL );
+			rs->sr_ref[0].bv_val ? rs->sr_ref[0].bv_val : "NULL" );
 	}
 
 #ifdef LDAP_CONNECTIONLESS
@@ -577,7 +754,7 @@ send_ldap_response(
 #endif
 		
 	if ( rc == -1 ) {
-		Debug( LDAP_DEBUG_ANY, "ber_printf failed\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_ANY, "ber_printf failed\n" );
 
 #ifdef LDAP_CONNECTIONLESS
 		if (!op->o_conn || op->o_conn->c_is_udp == 0)
@@ -599,8 +776,7 @@ send_ldap_response(
 
 	if ( bytes < 0 ) {
 		Debug( LDAP_DEBUG_ANY,
-			"send_ldap_response: ber write failed\n",
-			0, 0, 0 );
+			"send_ldap_response: ber write failed\n" );
 
 		goto cleanup;
 	}
@@ -657,15 +833,18 @@ send_ldap_disconnect( Operation	*op, SlapReply *rs )
 	|| (e) == LDAP_STRONG_AUTH_REQUIRED \
 	|| (e) == LDAP_UNAVAILABLE )
 
+	Debug( LDAP_DEBUG_TRACE,
+		"send_ldap_disconnect %d:%s\n",
+		rs->sr_err, rs->sr_text ? rs->sr_text : "" );
 	assert( LDAP_UNSOLICITED_ERROR( rs->sr_err ) );
 
+	/* TODO: Flush the entry if sr_type == REP_SEARCH/REP_SEARCHREF? */
+	RS_ASSERT( !(rs->sr_flags & REP_ENTRY_MASK) );
+	rs->sr_flags &= ~REP_ENTRY_MASK;	/* paranoia */
+
 	rs->sr_type = REP_EXTENDED;
 	rs->sr_rspdata = NULL;
 
-	Debug( LDAP_DEBUG_TRACE,
-		"send_ldap_disconnect %d:%s\n",
-		rs->sr_err, rs->sr_text ? rs->sr_text : "", NULL );
-
 	if ( op->o_protocol < LDAP_VERSION3 ) {
 		rs->sr_rspoid = NULL;
 		rs->sr_tag = slap_req2res( op->o_tag );
@@ -678,10 +857,11 @@ send_ldap_disconnect( Operation	*op, SlapReply *rs )
 	}
 
 	if ( send_ldap_response( op, rs ) == SLAP_CB_CONTINUE ) {
-		Statslog( LDAP_DEBUG_STATS,
-			"%s DISCONNECT tag=%lu err=%d text=%s\n",
+		ETIME_SETUP;
+		StatslogEtime( LDAP_DEBUG_STATS,
+			"%s DISCONNECT tag=%lu err=%d "ETIME_LOGFMT"text=%s\n",
 			op->o_log_prefix, rs->sr_tag, rs->sr_err,
-			rs->sr_text ? rs->sr_text : "", 0 );
+			rs->sr_text ? rs->sr_text : "" );
 	}
 }
 
@@ -698,25 +878,19 @@ slap_send_ldap_result( Operation *op, SlapReply *rs )
 	if ( rs->sr_err == SLAPD_ABANDON || op->o_abandon )
 		goto abandon;
 
-	assert( !LDAP_API_ERROR( rs->sr_err ) );
-
 	Debug( LDAP_DEBUG_TRACE,
 		"send_ldap_result: %s p=%d\n",
-		op->o_log_prefix, op->o_protocol, 0 );
-
+		op->o_log_prefix, op->o_protocol );
 	Debug( LDAP_DEBUG_ARGS,
 		"send_ldap_result: err=%d matched=\"%s\" text=\"%s\"\n",
 		rs->sr_err, rs->sr_matched ? rs->sr_matched : "",
 		rs->sr_text ? rs->sr_text : "" );
-
-
 	if( rs->sr_ref ) {
 		Debug( LDAP_DEBUG_ARGS,
 			"send_ldap_result: referral=\"%s\"\n",
-			rs->sr_ref[0].bv_val ? rs->sr_ref[0].bv_val : "NULL",
-			NULL, NULL );
+			rs->sr_ref[0].bv_val ? rs->sr_ref[0].bv_val : "NULL" );
 	}
-
+	assert( !LDAP_API_ERROR( rs->sr_err ) );
 	assert( rs->sr_err != LDAP_PARTIAL_RESULTS );
 
 	if ( rs->sr_err == LDAP_REFERRAL ) {
@@ -748,16 +922,17 @@ abandon:
 	}
 
 	if ( send_ldap_response( op, rs ) == SLAP_CB_CONTINUE ) {
+		ETIME_SETUP;
 		if ( op->o_tag == LDAP_REQ_SEARCH ) {
-			Statslog( LDAP_DEBUG_STATS,
-				"%s SEARCH RESULT tag=%lu err=%d nentries=%d text=%s\n",
+			StatslogEtime( LDAP_DEBUG_STATS,
+				"%s SEARCH RESULT tag=%lu err=%d "ETIME_LOGFMT"nentries=%d text=%s\n",
 				op->o_log_prefix, rs->sr_tag, rs->sr_err,
 				rs->sr_nentries, rs->sr_text ? rs->sr_text : "" );
 		} else {
-			Statslog( LDAP_DEBUG_STATS,
-				"%s RESULT tag=%lu err=%d text=%s\n",
+			StatslogEtime( LDAP_DEBUG_STATS,
+				"%s RESULT tag=%lu err=%d "ETIME_LOGFMT"text=%s\n",
 				op->o_log_prefix, rs->sr_tag, rs->sr_err,
-				rs->sr_text ? rs->sr_text : "", 0 );
+				rs->sr_text ? rs->sr_text : "" );
 		}
 	}
 
@@ -769,63 +944,77 @@ abandon:
 void
 send_ldap_sasl( Operation *op, SlapReply *rs )
 {
-	rs->sr_type = REP_SASL;
 	Debug( LDAP_DEBUG_TRACE, "send_ldap_sasl: err=%d len=%ld\n",
 		rs->sr_err,
-		rs->sr_sasldata ? (long) rs->sr_sasldata->bv_len : -1, NULL );
+		rs->sr_sasldata ? (long) rs->sr_sasldata->bv_len : -1 );
 
+	RS_ASSERT( !(rs->sr_flags & REP_ENTRY_MASK) );
+	rs->sr_flags &= ~REP_ENTRY_MASK;	/* paranoia */
+
+	rs->sr_type = REP_SASL;
 	rs->sr_tag = slap_req2res( op->o_tag );
 	rs->sr_msgid = (rs->sr_tag != LBER_SEQUENCE) ? op->o_msgid : 0;
 
 	if ( send_ldap_response( op, rs ) == SLAP_CB_CONTINUE ) {
-		Statslog( LDAP_DEBUG_STATS,
-			"%s RESULT tag=%lu err=%d text=%s\n",
+		ETIME_SETUP;
+		StatslogEtime( LDAP_DEBUG_STATS,
+			"%s RESULT tag=%lu err=%d "ETIME_LOGFMT"text=%s\n",
 			op->o_log_prefix, rs->sr_tag, rs->sr_err,
-			rs->sr_text ? rs->sr_text : "", 0 );
+			rs->sr_text ? rs->sr_text : "" );
 	}
 }
 
 void
 slap_send_ldap_extended( Operation *op, SlapReply *rs )
 {
-	rs->sr_type = REP_EXTENDED;
-
 	Debug( LDAP_DEBUG_TRACE,
 		"send_ldap_extended: err=%d oid=%s len=%ld\n",
 		rs->sr_err,
 		rs->sr_rspoid ? rs->sr_rspoid : "",
 		rs->sr_rspdata != NULL ? rs->sr_rspdata->bv_len : 0 );
 
+	RS_ASSERT( !(rs->sr_flags & REP_ENTRY_MASK) );
+	rs->sr_flags &= ~REP_ENTRY_MASK;	/* paranoia */
+
+	rs->sr_type = REP_EXTENDED;
 	rs->sr_tag = slap_req2res( op->o_tag );
 	rs->sr_msgid = (rs->sr_tag != LBER_SEQUENCE) ? op->o_msgid : 0;
 
 	if ( send_ldap_response( op, rs ) == SLAP_CB_CONTINUE ) {
-		Statslog( LDAP_DEBUG_STATS,
-			"%s RESULT oid=%s err=%d text=%s\n",
+		ETIME_SETUP;
+		StatslogEtime( LDAP_DEBUG_STATS,
+			"%s RESULT oid=%s err=%d "ETIME_LOGFMT"text=%s\n",
 			op->o_log_prefix, rs->sr_rspoid ? rs->sr_rspoid : "",
-			rs->sr_err, rs->sr_text ? rs->sr_text : "", 0 );
+			rs->sr_err, rs->sr_text ? rs->sr_text : "" );
 	}
 }
 
 void
 slap_send_ldap_intermediate( Operation *op, SlapReply *rs )
 {
-	rs->sr_type = REP_INTERMEDIATE;
 	Debug( LDAP_DEBUG_TRACE,
 		"send_ldap_intermediate: err=%d oid=%s len=%ld\n",
 		rs->sr_err,
 		rs->sr_rspoid ? rs->sr_rspoid : "",
 		rs->sr_rspdata != NULL ? rs->sr_rspdata->bv_len : 0 );
+
+	RS_ASSERT( !(rs->sr_flags & REP_ENTRY_MASK) );
+	rs->sr_flags &= ~REP_ENTRY_MASK;	/* paranoia */
+
+	rs->sr_type = REP_INTERMEDIATE;
 	rs->sr_tag = LDAP_RES_INTERMEDIATE;
 	rs->sr_msgid = op->o_msgid;
 	if ( send_ldap_response( op, rs ) == SLAP_CB_CONTINUE ) {
-		Statslog( LDAP_DEBUG_STATS2,
-			"%s INTERM oid=%s\n",
+		Debug( LDAP_DEBUG_STATS2,
+			"%s INTERMEDIATE oid=%s\n",
 			op->o_log_prefix,
-			rs->sr_rspoid ? rs->sr_rspoid : "", 0, 0, 0 );
+			rs->sr_rspoid ? rs->sr_rspoid : "" );
 	}
 }
 
+#define set_ldap_error( rs, err, text ) do { \
+		(rs)->sr_err = err; (rs)->sr_text = text; } while(0)
+
 /*
  * returns:
  *
@@ -854,19 +1043,21 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 	 */
 	char **e_flags = NULL;
 
+	rs->sr_type = REP_SEARCH;
+
 	if ( op->ors_slimit >= 0 && rs->sr_nentries >= op->ors_slimit ) {
-		return LDAP_SIZELIMIT_EXCEEDED;
+		rc = LDAP_SIZELIMIT_EXCEEDED;
+		goto error_return;
 	}
 
 	/* Every 64 entries, check for thread pool pause */
 	if ( ( ( rs->sr_nentries & 0x3f ) == 0x3f ) &&
 		ldap_pvt_thread_pool_pausing( &connection_pool ) > 0 )
 	{
-		return LDAP_BUSY;
+		rc = LDAP_BUSY;
+		goto error_return;
 	}
 
-	rs->sr_type = REP_SEARCH;
-
 	/* eventually will loop through generated operational attribute types
 	 * currently implemented types include:
 	 *	entryDN, subschemaSubentry, and hasSubordinates */
@@ -899,7 +1090,7 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 	if ( !access_allowed( op, rs->sr_entry, ad_entry, NULL, ACL_READ, NULL )) {
 		Debug( LDAP_DEBUG_ACL,
 			"send_search_entry: conn %lu access to entry (%s) not allowed\n", 
-			op->o_connid, rs->sr_entry->e_name.bv_val, 0 );
+			op->o_connid, rs->sr_entry->e_name.bv_val );
 
 		rc = LDAP_INSUFFICIENT_ACCESS;
 		goto error_return;
@@ -932,7 +1123,8 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 #endif
 	if ( op->o_res_ber ) {
 		/* read back control */
-	    rc = ber_printf( ber, "{O{" /*}}*/, &rs->sr_entry->e_name );
+	    rc = ber_printf( ber, "t{O{" /*}}*/,
+			LDAP_RES_SEARCH_ENTRY, &rs->sr_entry->e_name );
 	} else {
 	    rc = ber_printf( ber, "{it{O{" /*}}}*/, op->o_msgid,
 			LDAP_RES_SEARCH_ENTRY, &rs->sr_entry->e_name );
@@ -941,10 +1133,10 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 	if ( rc == -1 ) {
 		Debug( LDAP_DEBUG_ANY, 
 			"send_search_entry: conn %lu  ber_printf failed\n", 
-			op->o_connid, 0, 0 );
+			op->o_connid );
 
 		if ( op->o_res_ber == NULL ) ber_free_buf( ber );
-		send_ldap_error( op, rs, LDAP_OTHER, "encoding DN error" );
+		set_ldap_error( rs, LDAP_OTHER, "encoding DN error" );
 		rc = rs->sr_err;
 		goto error_return;
 	}
@@ -971,10 +1163,10 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 			if( e_flags == NULL ) {
 		    	Debug( LDAP_DEBUG_ANY, 
 					"send_search_entry: conn %lu slap_sl_calloc failed\n",
-					op->o_connid, 0, 0 );
+					op->o_connid );
 				ber_free( ber, 1 );
 	
-				send_ldap_error( op, rs, LDAP_OTHER, "out of memory" );
+				set_ldap_error( rs, LDAP_OTHER, "out of memory" );
 				goto error_return;
 			}
 			a_flags = (char *)(e_flags + i);
@@ -989,9 +1181,9 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 			if ( rc == -1 ) {
 			    	Debug( LDAP_DEBUG_ANY, "send_search_entry: "
 					"conn %lu matched values filtering failed\n",
-					op->o_connid, 0, 0 );
+					op->o_connid );
 				if ( op->o_res_ber == NULL ) ber_free_buf( ber );
-				send_ldap_error( op, rs, LDAP_OTHER,
+				set_ldap_error( rs, LDAP_OTHER,
 					"matched values filtering error" );
 				rc = rs->sr_err;
 				goto error_return;
@@ -1035,17 +1227,17 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 			{
 				Debug( LDAP_DEBUG_ACL, "send_search_entry: "
 					"conn %lu access to attribute %s not allowed\n",
-				        op->o_connid, desc->ad_cname.bv_val, 0 );
+				        op->o_connid, desc->ad_cname.bv_val );
 				continue;
 			}
 
 			if (( rc = ber_printf( ber, "{O[" /*]}*/ , &desc->ad_cname )) == -1 ) {
 				Debug( LDAP_DEBUG_ANY, 
 					"send_search_entry: conn %lu  ber_printf failed\n", 
-					op->o_connid, 0, 0 );
+					op->o_connid );
 
 				if ( op->o_res_ber == NULL ) ber_free_buf( ber );
-				send_ldap_error( op, rs, LDAP_OTHER,
+				set_ldap_error( rs, LDAP_OTHER,
 					"encoding description error");
 				rc = rs->sr_err;
 				goto error_return;
@@ -1076,10 +1268,10 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 					if (( rc = ber_printf( ber, "{O[" /*]}*/ , &desc->ad_cname )) == -1 ) {
 						Debug( LDAP_DEBUG_ANY,
 							"send_search_entry: conn %lu  ber_printf failed\n", 
-							op->o_connid, 0, 0 );
+							op->o_connid );
 
 						if ( op->o_res_ber == NULL ) ber_free_buf( ber );
-						send_ldap_error( op, rs, LDAP_OTHER,
+						set_ldap_error( rs, LDAP_OTHER,
 							"encoding description error");
 						rc = rs->sr_err;
 						goto error_return;
@@ -1088,10 +1280,10 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 				if (( rc = ber_printf( ber, "O", &a->a_vals[i] )) == -1 ) {
 					Debug( LDAP_DEBUG_ANY,
 						"send_search_entry: conn %lu  "
-						"ber_printf failed.\n", op->o_connid, 0, 0 );
+						"ber_printf failed.\n", op->o_connid );
 
 					if ( op->o_res_ber == NULL ) ber_free_buf( ber );
-					send_ldap_error( op, rs, LDAP_OTHER,
+					set_ldap_error( rs, LDAP_OTHER,
 						"encoding values error" );
 					rc = rs->sr_err;
 					goto error_return;
@@ -1102,10 +1294,10 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 		if ( finish && ( rc = ber_printf( ber, /*{[*/ "]N}" )) == -1 ) {
 			Debug( LDAP_DEBUG_ANY,
 				"send_search_entry: conn %lu ber_printf failed\n", 
-				op->o_connid, 0, 0 );
+				op->o_connid );
 
 			if ( op->o_res_ber == NULL ) ber_free_buf( ber );
-			send_ldap_error( op, rs, LDAP_OTHER, "encode end error" );
+			set_ldap_error( rs, LDAP_OTHER, "encode end error" );
 			rc = rs->sr_err;
 			goto error_return;
 		}
@@ -1136,9 +1328,9 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 					"send_search_entry: conn %lu "
 					"not enough memory "
 					"for matched values filtering\n",
-					op->o_connid, 0, 0 );
+					op->o_connid );
 				if ( op->o_res_ber == NULL ) ber_free_buf( ber );
-				send_ldap_error( op, rs, LDAP_OTHER,
+				set_ldap_error( rs, LDAP_OTHER,
 					"not enough memory for matched values filtering" );
 				goto error_return;
 			}
@@ -1156,9 +1348,9 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 			    	Debug( LDAP_DEBUG_ANY,
 					"send_search_entry: conn %lu "
 					"matched values filtering failed\n", 
-					op->o_connid, 0, 0);
+					op->o_connid );
 				if ( op->o_res_ber == NULL ) ber_free_buf( ber );
-				send_ldap_error( op, rs, LDAP_OTHER,
+				set_ldap_error( rs, LDAP_OTHER,
 					"matched values filtering error" );
 				rc = rs->sr_err;
 				goto error_return;
@@ -1183,6 +1375,10 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 				{
 					continue;
 				}
+				/* if DSA-specific and replicating, skip */
+				if ( op->o_sync != SLAP_CONTROL_NONE &&
+					desc->ad_type->sat_usage == LDAP_SCHEMA_DSA_OPERATION )
+					continue;
 			} else {
 				if ( !userattrs && !ad_inlist( desc, rs->sr_attrs ) ) {
 					continue;
@@ -1196,7 +1392,7 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 			Debug( LDAP_DEBUG_ACL,
 				"send_search_entry: conn %lu "
 				"access to attribute %s not allowed\n",
-				op->o_connid, desc->ad_cname.bv_val, 0 );
+				op->o_connid, desc->ad_cname.bv_val );
 
 			continue;
 		}
@@ -1205,10 +1401,10 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 		if ( rc == -1 ) {
 			Debug( LDAP_DEBUG_ANY,
 				"send_search_entry: conn %lu  "
-				"ber_printf failed\n", op->o_connid, 0, 0 );
+				"ber_printf failed\n", op->o_connid );
 
 			if ( op->o_res_ber == NULL ) ber_free_buf( ber );
-			send_ldap_error( op, rs, LDAP_OTHER,
+			set_ldap_error( rs, LDAP_OTHER,
 				"encoding description error" );
 			rc = rs->sr_err;
 			goto error_return;
@@ -1234,10 +1430,10 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 				if (( rc = ber_printf( ber, "O", &a->a_vals[i] )) == -1 ) {
 					Debug( LDAP_DEBUG_ANY,
 						"send_search_entry: conn %lu  ber_printf failed\n", 
-						op->o_connid, 0, 0 );
+						op->o_connid );
 
 					if ( op->o_res_ber == NULL ) ber_free_buf( ber );
-					send_ldap_error( op, rs, LDAP_OTHER,
+					set_ldap_error( rs, LDAP_OTHER,
 						"encoding values error" );
 					rc = rs->sr_err;
 					goto error_return;
@@ -1248,10 +1444,10 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 		if (( rc = ber_printf( ber, /*{[*/ "]N}" )) == -1 ) {
 			Debug( LDAP_DEBUG_ANY,
 				"send_search_entry: conn %lu  ber_printf failed\n",
-				op->o_connid, 0, 0 );
+				op->o_connid );
 
 			if ( op->o_res_ber == NULL ) ber_free_buf( ber );
-			send_ldap_error( op, rs, LDAP_OTHER, "encode end error" );
+			set_ldap_error( rs, LDAP_OTHER, "encode end error" );
 			rc = rs->sr_err;
 			goto error_return;
 		}
@@ -1283,22 +1479,18 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 	}
 
 	if ( rc == -1 ) {
-		Debug( LDAP_DEBUG_ANY, "ber_printf failed\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_ANY, "ber_printf failed\n" );
 
 		if ( op->o_res_ber == NULL ) ber_free_buf( ber );
-		send_ldap_error( op, rs, LDAP_OTHER, "encode entry end error" );
+		set_ldap_error( rs, LDAP_OTHER, "encode entry end error" );
 		rc = rs->sr_err;
 		goto error_return;
 	}
 
-	Statslog( LDAP_DEBUG_STATS2, "%s ENTRY dn=\"%s\"\n",
-	    op->o_log_prefix, rs->sr_entry->e_nname.bv_val, 0, 0, 0 );
+	Debug( LDAP_DEBUG_STATS2, "%s ENTRY dn=\"%s\"\n",
+	    op->o_log_prefix, rs->sr_entry->e_nname.bv_val );
 
-	if ( rs->sr_flags & REP_ENTRY_MUSTRELEASE ) {
-		be_entry_release_rw( op, rs->sr_entry, 0 );
-		rs->sr_flags ^= REP_ENTRY_MUSTRELEASE;
-		rs->sr_entry = NULL;
-	}
+	rs_flush_entry( op, rs, NULL );
 
 	if ( op->o_res_ber == NULL ) {
 		bytes = send_ldap_ber( op, ber );
@@ -1307,7 +1499,7 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 		if ( bytes < 0 ) {
 			Debug( LDAP_DEBUG_ANY,
 				"send_search_entry: conn %lu  ber write failed.\n", 
-				op->o_connid, 0, 0 );
+				op->o_connid );
 
 			rc = LDAP_UNAVAILABLE;
 			goto error_return;
@@ -1322,7 +1514,7 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 	}
 
 	Debug( LDAP_DEBUG_TRACE,
-		"<= send_search_entry: conn %lu exit.\n", op->o_connid, 0, 0 );
+		"<= send_search_entry: conn %lu exit.\n", op->o_connid );
 
 	rc = LDAP_SUCCESS;
 
@@ -1335,25 +1527,25 @@ error_return:;
 		slap_sl_free( e_flags, op->o_tmpmemctx );
 	}
 
+	/* FIXME: Can break if rs now contains an extended response */
 	if ( rs->sr_operational_attrs ) {
 		attrs_free( rs->sr_operational_attrs );
 		rs->sr_operational_attrs = NULL;
 	}
 	rs->sr_attr_flags = SLAP_ATTRS_UNDEFINED;
 
-	/* FIXME: I think rs->sr_type should be explicitly set to
-	 * REP_SEARCH here. That's what it was when we entered this
-	 * function. send_ldap_error may have changed it, but we
-	 * should set it back so that the cleanup functions know
-	 * what they're doing.
-	 */
-	if ( op->o_tag == LDAP_REQ_SEARCH && rs->sr_type == REP_SEARCH 
-		&& rs->sr_entry 
-		&& ( rs->sr_flags & REP_ENTRY_MUSTBEFREED ) ) 
-	{
-		entry_free( rs->sr_entry );
-		rs->sr_entry = NULL;
-		rs->sr_flags &= ~REP_ENTRY_MUSTBEFREED;
+	if ( op->o_tag == LDAP_REQ_SEARCH && rs->sr_type == REP_SEARCH ) {
+		rs_flush_entry( op, rs, NULL );
+	} else {
+		RS_ASSERT( (rs->sr_flags & REP_ENTRY_MASK) == 0 );
+	}
+
+	if ( rs->sr_flags & REP_CTRLS_MUSTBEFREED ) {
+		rs->sr_flags ^= REP_CTRLS_MUSTBEFREED; /* paranoia */
+		if ( rs->sr_ctrls ) {
+			slap_free_ctrls( op, rs->sr_ctrls );
+			rs->sr_ctrls = NULL;
+		}
 	}
 
 	return( rc );
@@ -1381,14 +1573,13 @@ slap_send_search_reference( Operation *op, SlapReply *rs )
 
 	Debug( LDAP_DEBUG_TRACE,
 		"=> send_search_reference: dn=\"%s\"\n",
-		edn, 0, 0 );
+		edn );
 
 	if (  rs->sr_entry && ! access_allowed( op, rs->sr_entry,
 		ad_entry, NULL, ACL_READ, NULL ) )
 	{
 		Debug( LDAP_DEBUG_ACL,
-			"send_search_reference: access to entry not allowed\n",
-		    0, 0, 0 );
+			"send_search_reference: access to entry not allowed\n" );
 		rc = 1;
 		goto rel;
 	}
@@ -1398,8 +1589,7 @@ slap_send_search_reference( Operation *op, SlapReply *rs )
 	{
 		Debug( LDAP_DEBUG_ACL,
 			"send_search_reference: access "
-			"to reference not allowed\n",
-		    0, 0, 0 );
+			"to reference not allowed\n" );
 		rc = 1;
 		goto rel;
 	}
@@ -1407,7 +1597,7 @@ slap_send_search_reference( Operation *op, SlapReply *rs )
 	if( op->o_domain_scope ) {
 		Debug( LDAP_DEBUG_ANY,
 			"send_search_reference: domainScope control in (%s)\n", 
-			edn, 0, 0 );
+			edn );
 		rc = 0;
 		goto rel;
 	}
@@ -1415,7 +1605,7 @@ slap_send_search_reference( Operation *op, SlapReply *rs )
 	if( rs->sr_ref == NULL ) {
 		Debug( LDAP_DEBUG_ANY,
 			"send_search_reference: null ref in (%s)\n", 
-			edn, 0, 0 );
+			edn );
 		rc = 1;
 		goto rel;
 	}
@@ -1453,23 +1643,18 @@ slap_send_search_reference( Operation *op, SlapReply *rs )
 
 	if ( rc == -1 ) {
 		Debug( LDAP_DEBUG_ANY,
-			"send_search_reference: ber_printf failed\n", 0, 0, 0 );
+			"send_search_reference: ber_printf failed\n" );
 
 #ifdef LDAP_CONNECTIONLESS
 		if (!op->o_conn || op->o_conn->c_is_udp == 0)
 #endif
 		ber_free_buf( ber );
-		send_ldap_error( op, rs, LDAP_OTHER, "encode DN error" );
+		set_ldap_error( rs, LDAP_OTHER, "encode DN error" );
 		goto rel;
 	}
 
 	rc = 0;
-	if ( rs->sr_flags & REP_ENTRY_MUSTRELEASE ) {
-		assert( rs->sr_entry != NULL );
-		be_entry_release_rw( op, rs->sr_entry, 0 );
-		rs->sr_flags ^= REP_ENTRY_MUSTRELEASE;
-		rs->sr_entry = NULL;
-	}
+	rs_flush_entry( op, rs, NULL );
 
 #ifdef LDAP_CONNECTIONLESS
 	if (!op->o_conn || op->o_conn->c_is_udp == 0) {
@@ -1493,23 +1678,34 @@ slap_send_search_reference( Operation *op, SlapReply *rs )
 		int	r;
 
 		for ( r = 0; !BER_BVISNULL( &rs->sr_ref[ r ] ); r++ ) {
-			Statslog( LDAP_DEBUG_STATS2, "%s REF #%d \"%s\"\n",
-				op->o_log_prefix, r, rs->sr_ref[0].bv_val,
-				0, 0 );
+			Debug( LDAP_DEBUG_STATS2, "%s REF #%d \"%s\"\n",
+				op->o_log_prefix, r, rs->sr_ref[0].bv_val );
 		}
 
 	} else {
-		Statslog( LDAP_DEBUG_STATS2, "%s REF \"(null)\"\n",
-			op->o_log_prefix, 0, 0, 0, 0 );
+		Debug( LDAP_DEBUG_STATS2, "%s REF \"(null)\"\n",
+			op->o_log_prefix );
 	}
 
-	Debug( LDAP_DEBUG_TRACE, "<= send_search_reference\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "<= send_search_reference\n" );
 
+	if ( 0 ) {
 rel:
+	    rs_flush_entry( op, rs, NULL );
+	}
+
 	if ( op->o_callback ) {
 		(void)slap_cleanup_play( op, rs );
 	}
 
+	if ( rs->sr_flags & REP_CTRLS_MUSTBEFREED ) {
+		rs->sr_flags ^= REP_CTRLS_MUSTBEFREED; /* paranoia */
+		if ( rs->sr_ctrls ) {
+			slap_free_ctrls( op, rs->sr_ctrls );
+			rs->sr_ctrls = NULL;
+		}
+	}
+
 	return rc;
 }
 
@@ -1529,7 +1725,7 @@ str2result(
 
 	if ( strncasecmp( s, "RESULT", STRLENOF( "RESULT" ) ) != 0 ) {
 		Debug( LDAP_DEBUG_ANY, "str2result (%s) expecting \"RESULT\"\n",
-		    s, 0, 0 );
+		    s );
 
 		return( -1 );
 	}
@@ -1550,7 +1746,7 @@ str2result(
 
 			if ( c == NULL ) {
 				Debug( LDAP_DEBUG_ANY, "str2result (%s) missing value\n",
-				    s, 0, 0 );
+				    s );
 				rc = -1;
 				continue;
 			}
@@ -1558,7 +1754,7 @@ str2result(
 			while ( isspace( (unsigned char) c[ 0 ] ) ) c++;
 			if ( c[ 0 ] == '\0' ) {
 				Debug( LDAP_DEBUG_ANY, "str2result (%s) missing or empty value\n",
-				    s, 0, 0 );
+				    s );
 				rc = -1;
 				continue;
 			}
@@ -1566,20 +1762,21 @@ str2result(
 			retcode = strtol( c, &next, 10 );
 			if ( next == NULL || next == c ) {
 				Debug( LDAP_DEBUG_ANY, "str2result (%s) unable to parse value\n",
-				    s, 0, 0 );
+				    s );
 				rc = -1;
 				continue;
 			}
 
-			while ( isspace( (unsigned char) next[ 0 ] ) ) next++;
-			if ( next[ 0 ] != '\0' ) {
+			while ( isspace( (unsigned char) next[ 0 ] ) && next[ 0 ] != '\n' )
+				next++;
+			if ( next[ 0 ] != '\0' && next[ 0 ] != '\n' ) {
 				Debug( LDAP_DEBUG_ANY, "str2result (%s) extra cruft after value\n",
-				    s, 0, 0 );
+				    s );
 				rc = -1;
 				continue;
 			}
 
-			/* FIXME: what if it's larger that max int? */
+			/* FIXME: what if it's larger than max int? */
 			*code = (int)retcode;
 
 		} else if ( strncasecmp( s, "matched", STRLENOF( "matched" ) ) == 0 ) {
@@ -1592,7 +1789,7 @@ str2result(
 			}
 		} else {
 			Debug( LDAP_DEBUG_ANY, "str2result (%s) unknown\n",
-			    s, 0, 0 );
+			    s );
 
 			rc = -1;
 		}
@@ -1635,6 +1832,7 @@ int slap_read_controls(
 	myop.o_res_ber = ber;
 	myop.o_callback = NULL;
 	myop.ors_slimit = 1;
+	myop.ors_attrsonly = 0;
 
 	rc = slap_send_search_entry( &myop, rs );
 	if( rc ) return rc;
