@@ -2,7 +2,7 @@
 /* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2001-2015 The OpenLDAP Foundation.
+ * Copyright 2001-2024 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -27,7 +27,7 @@
 
 #include "slap.h"
 #include "slapi.h"
-#include "config.h"
+#include "slap-config.h"
 
 #ifdef LDAP_SLAPI
 
@@ -82,6 +82,7 @@ slapi_op_internal_p( Operation *op, SlapReply *rs, slap_callback *cb )
 		cb->sc_response = slapi_over_response;
 		cb->sc_cleanup = slapi_over_cleanup;
 		cb->sc_private = pb;
+		cb->sc_writewait = 0;
 		cb->sc_next = op->o_callback;
 		op->o_callback = cb;
 	}
@@ -276,13 +277,13 @@ slapi_op_bind_callback( Operation *op, SlapReply *rs, int prc )
 			ldap_pvt_thread_mutex_unlock( &op->o_conn->c_mutex );
 
 			/* log authorization identity */
-			Statslog( LDAP_DEBUG_STATS,
+			Debug( LDAP_DEBUG_STATS,
 				"%s BIND dn=\"%s\" mech=%s (SLAPI) ssf=0\n",
 				op->o_log_prefix,
 				BER_BVISNULL( &op->o_conn->c_dn )
 					? "<empty>" : op->o_conn->c_dn.bv_val,
 				BER_BVISNULL( &op->orb_mech )
-					? "<empty>" : op->orb_mech.bv_val, 0, 0 );
+					? "<empty>" : op->orb_mech.bv_val );
 
 			return -1;
 		}
@@ -734,6 +735,7 @@ slapi_over_access_allowed(
 
 	cb.sc_response = NULL;
 	cb.sc_cleanup = NULL;
+	cb.sc_writewait = NULL;
 
 	pb = SLAPI_OPERATION_PBLOCK( op );
 
@@ -798,6 +800,7 @@ slapi_over_acl_group(
 
 		cb.sc_response = NULL;
 		cb.sc_cleanup = NULL;
+		cb.sc_writewait = NULL;
 
 		pb = SLAPI_OPERATION_PBLOCK( op );
 
@@ -924,7 +927,7 @@ int slapi_over_config( BackendDB *be, ConfigReply *cr )
 	if ( slapi_over_initialized == 0 ) {
 		int rc;
 
-		/* do global initializaiton */
+		/* do global initialization */
 		ldap_pvt_thread_mutex_init( &slapi_hn_mutex );
 		ldap_pvt_thread_mutex_init( &slapi_time_mutex );
 		ldap_pvt_thread_mutex_init( &slapi_printmessage_mutex );
